import torch
import pytest
from tests.constants import (
    TOLERANCES,
    MILD_TOLERANCES,
    LOGM2_MEAN_STD,
    REPS,
    FRAMES_PREDICTOR,
)
from tests.helpers import sample_particle, equivectors_builder
from tests.hep import get_tagging_features

from lloca.backbone.particlenet import EdgeConvBlock, ParticleNet
from lloca.reps.tensorreps import TensorReps
from lloca.reps.tensorreps_transform import TensorRepsTransform
from lloca.utils.rand_transforms import rand_lorentz
from lloca.framesnet.frames import InverseFrames
from lloca.framesnet.equi_frames import LearnedSO13Frames, LearnedPDFrames


@pytest.mark.parametrize("FramesPredictor", FRAMES_PREDICTOR)
@pytest.mark.parametrize("batch_dims", [[10]])
@pytest.mark.parametrize("k", [2, 5])
@pytest.mark.parametrize("out_feats", [(12, 12)])
@pytest.mark.parametrize("hidden_reps", REPS)
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_edgeconvblock_invariance_equivariance(
    FramesPredictor,
    batch_dims,
    k,
    out_feats,
    logm2_std,
    logm2_mean,
    hidden_reps,
):
    dtype = torch.float64

    assert len(batch_dims) == 1
    equivectors = equivectors_builder()
    predictor = FramesPredictor(equivectors=equivectors).to(dtype=dtype)
    call_predictor = lambda fm: predictor(fm)

    # define edgeconv
    in_reps = TensorReps("1x1n")
    hidden_reps = TensorReps(hidden_reps)
    trafo = TensorRepsTransform(TensorReps(in_reps))
    linear_in = torch.nn.Linear(in_reps.dim, hidden_reps.dim).to(dtype=dtype)
    linear_out = torch.nn.Linear(out_feats[-1], in_reps.dim).to(dtype=dtype)
    edgeconv = EdgeConvBlock(k=k, in_reps=hidden_reps, out_feats=out_feats).to(dtype)

    def edgeconvblock_wrapper(x, frames):
        # use features as points for simplicity
        # this is equivariant, because features in local frame are invariant
        # and hence the knn ordering on them is also invariant
        # have to reshape to match ParticleNet format
        x = x.transpose(-1, -2).unsqueeze(0)
        x = edgeconv(points=x, features=x, frames=frames)
        x = x.transpose(-1, -2).squeeze(0)
        return x

    # get global transformation
    random = rand_lorentz([1], dtype=dtype)
    random = random.repeat(*batch_dims, 1, 1)

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)
    frames = call_predictor(fm)
    fm_local = trafo(fm, frames)

    # edgeconv - global
    x_local = linear_in(fm_local)
    x_prime_local = edgeconvblock_wrapper(x_local, frames)
    fm_prime_local = linear_out(x_prime_local)
    # back to global
    fm_prime_global = trafo(fm_prime_local, InverseFrames(frames))
    fm_prime_tr_global = torch.einsum("...ij,...j->...i", random, fm_prime_global)

    # global - edgeconv
    fm_transformed = torch.einsum("...ij,...j->...i", random, fm)
    frames_transformed = call_predictor(fm_transformed)
    fm_tr_local = trafo(fm_transformed, frames_transformed)
    x_tr_local = linear_in(fm_tr_local)
    x_tr_prime_local = edgeconvblock_wrapper(x_tr_local, frames_transformed)
    fm_tr_prime_local = linear_out(x_tr_prime_local)
    # back to global frame
    fm_tr_prime_global = trafo(fm_tr_prime_local, InverseFrames(frames_transformed))

    # test feature invariance before the operation
    torch.testing.assert_close(x_local, x_tr_local, **TOLERANCES)

    # test feature invariance after the operation
    torch.testing.assert_close(x_tr_prime_local, x_prime_local, **TOLERANCES)

    # test equivariance of outputs
    torch.testing.assert_close(fm_tr_prime_global, fm_prime_tr_global, **TOLERANCES)


@pytest.mark.parametrize(
    "FramesPredictor", [LearnedSO13Frames, LearnedPDFrames]
)  # RestFrames gives nans sometimes
@pytest.mark.parametrize("batch_dims", [[10]])
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_particlenet_invariance(
    FramesPredictor,
    batch_dims,
    logm2_std,
    logm2_mean,
):
    dtype = torch.float64
    batch = torch.zeros(batch_dims[0], dtype=torch.long)

    assert len(batch_dims) == 1
    equivectors = equivectors_builder()
    predictor = FramesPredictor(equivectors=equivectors).to(dtype=dtype)
    call_predictor = lambda fm: predictor(fm)

    # define particlenet
    in_reps = TensorReps("1x1n")
    trafo = TensorRepsTransform(TensorReps(in_reps))
    hidden_reps_list = ["3x0n+1x1n", "16x0n+4x1n"]  # pick something
    particlenet = ParticleNet(
        input_dims=TensorReps(hidden_reps_list[0]).dim,
        hidden_reps_list=hidden_reps_list,
        num_classes=1,
    ).to(dtype=dtype)
    particlenet.eval()  # turn off dropout

    def edgeconvblock_wrapper(p_local, frames):
        fts_local = get_tagging_features(p_local, batch)
        fts_local = fts_local.transpose(-1, -2).unsqueeze(0)
        points_local = fts_local[:, [4, 5], :]
        x = particlenet(points=points_local, features=fts_local, frames=frames)
        x = x.transpose(-1, -2).squeeze(0)
        return x

    # get global transformation
    random = rand_lorentz([1], dtype=dtype)
    random = random.repeat(*batch_dims, 1, 1)

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)
    frames = call_predictor(fm)
    fm_local = trafo(fm, frames)

    # particlenet
    score_prime_local = edgeconvblock_wrapper(fm_local, frames)

    # global - particlenet
    fm_transformed = torch.einsum("...ij,...j->...i", random, fm)
    frames_transformed = call_predictor(fm_transformed)
    fm_tr_local = trafo(fm_transformed, frames_transformed)
    score_tr_prime_local = edgeconvblock_wrapper(fm_tr_local, frames_transformed)

    # test feature invariance before the operation
    torch.testing.assert_close(fm_local, fm_tr_local, **TOLERANCES)

    # test equivariance of scores
    torch.testing.assert_close(
        score_tr_prime_local, score_prime_local, **MILD_TOLERANCES
    )



import torch
import pytest
from tests.constants import TOLERANCES, BATCH_DIMS

from lloca.utils.lorentz import lorentz_cross, lorentz_inner


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_lorentz_cross(batch_dims):
    # cross product of 3 random vectors
    v1 = torch.randn(batch_dims + [4])
    v2 = torch.randn(batch_dims + [4])
    v3 = torch.randn(batch_dims + [4])
    v4 = lorentz_cross(v1, v2, v3)

    # compute inner product of 4th vector with the first 3
    inner14 = lorentz_inner(v1, v4)
    inner24 = lorentz_inner(v2, v4)
    inner34 = lorentz_inner(v3, v4)

    # check that the inner products vanish
    zeros = torch.zeros_like(inner14)
    torch.testing.assert_close(inner14, zeros, **TOLERANCES)
    torch.testing.assert_close(inner24, zeros, **TOLERANCES)
    torch.testing.assert_close(inner34, zeros, **TOLERANCES)



import torch


def orthogonalize_3d(
    vecs, method="gramschmidt", eps_norm=1e-15, eps_reg=1e-10, return_reg=False
):
    """Wrapper for orthogonalization of euclidean vectors.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List of torch.tensor of shape (..., 3)
        Vectors to be orthogonalized
    method : str
        Method for orthogonalization. Options are "cross" and "gramschmidt".
    eps_norm : float
        Numerical regularization for the normalization of the vectors.
    eps_reg : float
        Controls the scale of the regularization for collinear vectors.
        eps_reg**2 defines the selection threshold.
    return_reg : bool
        If True, return a tuple with the orthogonalized vectors and the number of
        regularized vectors for collinearity.

    Returns
    -------
    orthogonal_vecs : list of torch.Tensor
        List of orthogonalized vectors of shape (..., 3)
    reg_collinear : int
        Number of vectors that were regularized due to collinearity.
    """
    vecs, reg_collinear = regularize_collinear(vecs, eps_reg)

    if method == "cross":
        trafo = orthogonalize_cross_3d(vecs, eps_norm)
    elif method == "gramschmidt":
        trafo = orthogonalize_gramschmidt_3d(vecs, eps_norm)
    else:
        raise ValueError(f"Orthogonalization method {method} not implemented")

    return (trafo, reg_collinear) if return_reg else trafo


def orthogonalize_gramschmidt_3d(vecs, eps_norm=1e-15):
    """Gram-Schmidt orthogonalization algorithm for O(3) vectors.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List of two vectors of shape (..., 3).
    eps_norm : float
        Numerical regularization for the normalization of the vectors.

    Returns
    -------
    orthogonal_vecs : list of torch.Tensor
        List of orthogonalized vectors of shape (..., 3).
    """
    n_vectors = len(vecs)
    assert n_vectors == 2

    vecs = [normalize_3d(v, eps_norm) for v in vecs]

    v_nexts = [v for v in vecs]
    orthogonal_vecs = [vecs[0]]

    # gram schmidt procedure
    for i in range(1, n_vectors):
        for k in range(i, n_vectors):
            v_inner = torch.sum(
                v_nexts[k] * orthogonal_vecs[i - 1], dim=-1, keepdim=True
            )
            v_nexts[k] = v_nexts[k] - orthogonal_vecs[i - 1] * v_inner
        orthogonal_vecs.append(normalize_3d(v_nexts[i], eps_norm))

    # last vector from cross product
    last_vec = torch.cross(*orthogonal_vecs, dim=-1)
    orthogonal_vecs.append(normalize_3d(last_vec, eps_norm))

    return orthogonal_vecs


def orthogonalize_cross_3d(vecs, eps_norm=1e-15):
    """Cross product orthogonalization algorithm for O(3) vectors.
    This is an alternative to the Gram-Schmidt algorithm, which is more numerically stable.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List of two vectors of shape (..., 3).
    eps_norm : float
        Numerical regularization for the normalization of the vectors.

    Returns
    -------
    orthogonal_vecs : list of torch.Tensor
        List of orthogonalized vectors of shape (..., 3).
    """
    n_vectors = len(vecs)
    assert n_vectors == 2

    vecs = [normalize_3d(v, eps_norm) for v in vecs]

    orthogonal_vecs = [vecs[0]]
    for i in range(1, n_vectors + 1):
        v_next = torch.cross(*orthogonal_vecs, *vecs[i:], dim=-1)
        assert torch.isfinite(v_next).all()
        orthogonal_vecs.append(normalize_3d(v_next, eps_norm))

    return orthogonal_vecs


def regularize_collinear(vecs, eps_reg=1e-10):
    """If the cross product of two vectors is small, the vectors are collinear.
    In this case, we add a small amount of noise to the second vector to
    regularize the orthogonalization.

    Parameters
    ----------
    vecs : list of torch.Tensor
        List with 2 vectors of shape (..., 3).
    eps_reg : float
        Regularization epsilon, controls the scale of the noise added to the second vector.

    Returns
    -------
    vecs : list of torch.Tensor
        List with 2 vectors of shape (..., 3), where the second vector is regularized if collinear.
    reg_collinear : int
        Number of vectors that were regularized due to collinearity.
    """
    assert len(vecs) == 2
    mask = torch.linalg.norm(torch.cross(*vecs, dim=-1)).abs() < eps_reg**2
    vecs[1][mask] += eps_reg * torch.randn_like(vecs[1][mask])

    reg_collinear = mask.sum().item()
    return vecs, reg_collinear


def normalize_3d(v, eps_norm=1e-15):
    """Normalize a 3D vector with numerical stability.

    Parameters
    ----------
    v : torch.Tensor
        A tensor of shape (..., 3) representing the vector to be normalized.
    eps_norm : float, optional
        A small value to prevent division by zero, by default 1e-10.

    Returns
    -------
    torch.Tensor
        The normalized vector of shape (..., 3).
    """
    norm = torch.linalg.norm(v, dim=-1, keepdim=True)
    return v / (norm + eps_norm)



"""PyTorch's modern and flexible flex_attention backend."""
try:
    from torch.nn.attention.flex_attention import flex_attention
except ModuleNotFoundError:
    raise ImportError(
        "torch>=2.5 is not installed. Run 'pip install lloca[flex_attention]'."
    )

attention = flex_attention



import torch

from ..framesnet.frames import Frames
from .tensorreps import TensorReps


class TensorRepsTransform(torch.nn.Module):
    def __init__(
        self,
        reps: TensorReps,
        use_naive=False,
    ):
        """Initialize tensor representation transformation module.

        Parameters
        ----------
        reps: TensorReps
            Tensor representations to transform.
        use_naive: bool
            Whether to use the naive transformation method.
            If False, uses an efficient transformation method.
            Default is False.
        """
        super().__init__()
        self.reps = reps
        self.transform = (
            self._transform_naive if use_naive else self._transform_efficient
        )

        # cache idx_start and idx_end for each rep
        self.start_end_idx = []
        idx = 0
        for mul_rep in self.reps:
            _, rep = mul_rep
            self.start_end_idx.append([idx, idx + mul_rep.dim])
            idx += mul_rep.dim

        # build parity_mask
        parity_odd = torch.zeros(self.reps.dim, dtype=torch.bool)
        idx = 0
        for mul_rep in self.reps:
            _, rep = mul_rep
            parity_odd[idx : idx + mul_rep.dim] = True if rep.parity == -1 else False
            idx += mul_rep.dim
        self.register_buffer("parity_odd", parity_odd.unsqueeze(0))
        self.no_parity_odd = self.parity_odd.sum() == 0

        if not use_naive:
            # build mapping from order to element in the reps list
            # only used in _transform_efficient
            self.map_rep = [None for _ in range(self.reps.max_rep.rep.order + 1)]
            idx_rep = 0
            for i in range(self.reps.max_rep.rep.order + 1):
                if reps[idx_rep].rep.order == i:
                    self.map_rep[i] = idx_rep
                    idx_rep += 1

        if self.reps.max_rep.rep.order <= 1:
            # super efficient shortcut if only scalar and vector reps are present
            self.transform = self._transform_only_scalars_and_vectors

    def forward(self, tensor: torch.Tensor, frames: Frames):
        """Apply a transformation to a tensor of a given representation.

        Parameters
        ----------
        tensor: torch.Tensor
            The tensor to transform, shape (..., self.reps.dim).
        frames: Frames
            The local frames to apply the transformation with, shape (..., 4, 4).

        Returns
        -------
        torch.Tensor
            The transformed tensor, shape (..., self.reps.dim).
        """
        assert self.reps.dim == tensor.shape[-1]

        if frames.is_identity or self.reps.mul_without_scalars == 0:
            return tensor

        in_shape = tensor.shape
        if len(frames.shape) > 3:
            frames = frames.reshape(-1, 4, 4)
        tensor = tensor.reshape(-1, tensor.shape[-1])
        assert tensor.shape[0] == frames.shape[0]

        tensor_transformed = self.transform(tensor, frames)
        tensor_transformed = self.transform_parity(tensor_transformed, frames)

        tensor_transformed = tensor_transformed.reshape(*in_shape)
        return tensor_transformed

    def _transform_naive(self, tensor, frames):
        """Naive transform: Apply n transformations to a tensor of n'th order.

        Parameters
        ----------
        tensor: torch.Tensor
            The tensor to transform, shape (N, self.reps.dim).
        frames: Frames
            The local frames to apply the transformation with, shape (N, 4, 4).

        Returns
        -------
        torch.Tensor
            The transformed tensor, shape (N, self.reps.dim).
        """
        output = tensor.clone()
        for mul_rep, [idx_start, idx_end] in zip(self.reps, self.start_end_idx):
            mul, rep = mul_rep
            if mul == 0 or rep.order == 0:
                continue

            x = tensor[:, idx_start:idx_end].reshape(-1, mul, *([4] * rep.order))

            einsum_string = get_einsum_string(rep.order)
            x_transformed = torch.einsum(
                einsum_string, *([frames.matrices.to(x.dtype)] * rep.order), x
            )
            output[:, idx_start:idx_end] = x_transformed.reshape(-1, mul_rep.dim)

        return output

    def _transform_efficient(self, tensor, frames):
        """Efficient transform:
        Starting with the highest-order tensor contribution,
        add the next contribution, apply frames transformation
        and flatten first dimension before continueing with next order.

        This is more efficient, because we use the
        maximum amount of parallelization possible.

        Parameters
        ----------
        tensor: torch.Tensor
            The tensor to transform, shape (N, self.reps.dim).
        frames: Frames
            The local frames to apply the transformation with, shape (N, 4, 4).

        Returns
        -------
        torch.Tensor
            The transformed tensor, shape (N, self.reps.dim).
        """
        output = None
        for order in reversed(range(self.reps.max_rep.rep.order + 1)):
            if self.map_rep[order] is not None:
                # add new contribution to the mix
                idx_start, idx_end = self.start_end_idx[self.map_rep[order]]
                contribution = tensor[:, idx_start:idx_end].reshape(
                    tensor.shape[0], -1, *(order * (4,))
                )
                output = (
                    torch.cat([contribution, output], dim=1)
                    if order < self.reps.max_rep.rep.order
                    else contribution
                )

            if order > 0:
                # apply transformation, then flatten because transformation is done
                output = torch.einsum(
                    "ijk,ilk...->ilj...", frames.matrices.to(output.dtype), output
                )
                output = output.flatten(start_dim=1, end_dim=2)

        return output

    def _transform_only_scalars_and_vectors(self, tensor, frames):
        """Super efficient transform that assumes that only scalars and vectors are present.
        Follows the same recipe as _transform_efficient, but avoids small overheads from
        torch.cat and torch.reshape.

        Parameters
        ----------
        tensor: torch.Tensor
            The tensor to transform, shape (N, self.reps.dim).
        frames: Frames
            The local frames to apply the transformation with, shape (N, 4, 4).

        Returns
        -------
        torch.Tensor
            The transformed tensor, shape (N, self.reps.dim).
        """
        output = tensor.clone()
        vector_idx_start, vector_idx_end = self.start_end_idx[-1]
        vectors = tensor[:, vector_idx_start:vector_idx_end]
        vectors = vectors.reshape(tensor.shape[0], -1, 4)
        vectors_transformed = torch.einsum(
            "ijk,ilk->ilj",
            frames.matrices.to(vectors.dtype),
            vectors,
        )
        output[:, vector_idx_start:vector_idx_end] = vectors_transformed.reshape(
            tensor.shape[0], -1
        )
        return output

    def transform_parity(self, tensor, frames):
        """Parity transform: Multiply parity-odd states by sign(det Lambda).

        Parameters
        ----------
        tensor: torch.Tensor
            The tensor to transform, shape (N, self.reps.dim).
        frames: Frames
            The local frames to apply the transformation with, shape (N, 4, 4).

        Returns
        -------
        torch.Tensor
            The transformed tensor, shape (N, self.reps.dim).
        """
        if self.no_parity_odd:
            return tensor
        else:
            return torch.where(
                self.parity_odd, frames.det.sign().unsqueeze(-1) * tensor, tensor
            )


def get_einsum_string(order):
    """Create einsum string for transformation of order-n tensor in _transform_naive.

    Parameters
    ----------
    order: int
        The order of the tensor representation.

    Returns
    -------
    str
        The einsum string for the transformation.
    """
    if order > 12:
        raise NotImplementedError("Running out of letters for order>12")

    einsum = ""
    start = ord("A")
    batch_index = ord("a")

    # list of frames
    for i in range(order):
        einsum += chr(batch_index) + chr(start + 2 * i) + chr(start + 2 * i + 1) + ","

    # tensor
    einsum += chr(batch_index)
    einsum += chr(start + 2 * order + 1)
    for i in range(order):
        einsum += chr(start + 2 * i + 1)

    # output
    einsum += "->"
    einsum += chr(batch_index)
    einsum += chr(start + 2 * order + 1)
    for i in range(order):
        einsum += chr(start + 2 * i)

    return einsum



import torch
from typing import List

from .lorentz import lorentz_eye
from .polar_decomposition import restframe_boost


def get_trafo_type(axis):
    """Determine whether the transformation is a boost or a rotation.

    Parameters
    ----------
    axis : torch.Tensor
        Axis of the transformation, should be of shape (2, *dims)

    Returns
    -------
    trafo_type : torch.Tensor
        A boolean tensor indicating whether the transformation is a boost (1) or a rotation (0).
    """
    return torch.any(axis == 0, dim=0)


def transform(
    axes: List[int],
    angles: List[torch.Tensor],
    use_float64=True,
):
    """Recursively build transformation matrices based on given lists of axes and angles

    Parameters
    ----------
    axes : List[torch.Tensor]
        List of axes along which the transformations are performed.
        Each element should be a tensor of shape (2, *dims).
    angles : List[torch.Tensor]
        List of angles used for the transformations.
        Each element should be a tensor of shape (*dims,).
    use_float64 : bool, optional
        Whether to use float64 for calculations, by default True

    Returns
    -------
    final_trafo : torch.Tensor
        Final transformation matrix of shape (*dims, 4, 4).
    """
    assert len(axes) == len(angles)
    dims = angles[0].shape
    assert all([angle.shape == dims for angle in angles])
    assert all([axis[0].shape == dims for axis in axes])
    assert all([axis[1].shape == dims for axis in axes])

    in_dtype = angles[0].dtype
    dtype = torch.float64 if use_float64 else in_dtype
    angles = [a.to(dtype=dtype) for a in angles]

    final_trafo = lorentz_eye(dims, angles[0].device, angles[0].dtype)
    for axis, angle in zip(axes, angles):
        trafo = lorentz_eye(dims, angle.device, angle.dtype)
        trafo_type = get_trafo_type(axis)

        meshgrid = torch.meshgrid(*[torch.arange(d) for d in dims], indexing="ij")
        trafo[(*meshgrid, axis[0], axis[0])] = torch.where(
            trafo_type, torch.cosh(angle), torch.cos(angle)
        )
        trafo[(*meshgrid, axis[0], axis[1])] = torch.where(
            trafo_type, torch.sinh(angle), -torch.sin(angle)
        )
        trafo[(*meshgrid, axis[1], axis[0])] = torch.where(
            trafo_type, torch.sinh(angle), torch.sin(angle)
        )
        trafo[(*meshgrid, axis[1], axis[1])] = torch.where(
            trafo_type, torch.cosh(angle), torch.cos(angle)
        )
        final_trafo = torch.einsum("...jk,...kl->...jl", trafo, final_trafo)
    return final_trafo.to(in_dtype)


def rand_lorentz(
    shape: List[int],
    std_eta: float = 0.1,
    n_max_std_eta: float = 3.0,
    device: str = "cpu",
    dtype: torch.dtype = torch.float32,
    generator: torch.Generator = None,
):
    """
    Create general Lorentz transformations as rotation * boost.

    Parameters
    ----------
    shape: List[int]
        Shape of the transformation matrices
    std_eta: float
        Standard deviation of rapidity
    n_max_std_eta: float
        Allowed number of standard deviations;
        used to sample from a truncated Gaussian
    device: str
    dtype: torch.dtype
    generator: torch.Generator

    Returns
    -------
    final_trafo: torch.tensor
        The resulting Lorentz transformation matrices of shape (*shape, 4, 4).
    """
    assert std_eta > 0
    boost = rand_boost(
        shape,
        std_eta,
        n_max_std_eta,
        device=device,
        dtype=dtype,
        generator=generator,
    )
    rotation = rand_rotation(shape, device, dtype, generator=generator)

    trafo = torch.einsum("...ij,...jk->...ik", rotation, boost)
    return trafo


def rand_xyrotation(
    shape: List[int],
    device: str = "cpu",
    dtype: torch.dtype = torch.float32,
    generator: torch.Generator = None,
):
    """
    Create xy-plane rotation matrices embedded in the Lorentz group.
    This function is a special case of rand_rotation.

    Parameters
    ----------
    shape: List[int]
        Shape of the transformation matrices
    device: str
    dtype: torch.dtype
    generator: torch.Generator

    Returns
    -------
    final_trafo: torch.tensor
        The resulting Lorentz transformation matrices of shape (*shape, 4, 4).
    """
    axis = torch.tensor([1, 2], dtype=torch.long, device=device)
    axis = axis.view(2, *([1] * len(shape))).repeat(1, *shape)
    angle = (
        torch.rand(*shape, device=device, dtype=dtype, generator=generator)
        * 2
        * torch.pi
    )
    return transform([axis], [angle])


def rand_ztransform(
    shape: List[int],
    std_eta: float = 0.1,
    n_max_std_eta: float = 3.0,
    device: str = "cpu",
    dtype: torch.dtype = torch.float32,
    generator: torch.Generator = None,
):
    """
    Create Lorentz transformations consisting of a boost along
    the z-axis and a rotation around the z-axis.
    This transformation is common in LHC physics.

    Parameters
    ----------
    shape: List[int]
        Shape of the transformation matrices
    std_eta: float
        Standard deviation of rapidity
    n_max_std_eta: float
        Allowed number of standard deviations;
        used to sample from a truncated Gaussian
    device: str
    dtype: torch.dtype
    generator: torch.Generator

    Returns
    -------
    final_trafo: torch.tensor
        The resulting Lorentz transformation matrices of shape (*shape, 4, 4).
    """
    # rotation around z-axis
    axis1 = torch.tensor([1, 2], dtype=torch.long, device=device)
    axis1 = axis1.view(2, *([1] * len(shape))).repeat(1, *shape)
    angle1 = (
        torch.rand(*shape, device=device, dtype=dtype, generator=generator)
        * 2
        * torch.pi
    )

    # boost along z-axis
    axis2 = torch.tensor([0, 3], dtype=torch.long, device=device)
    axis2 = axis2.view(2, *([1] * len(shape))).repeat(1, *shape)
    angle2 = sample_rapidity(
        shape,
        std_eta=std_eta,
        n_max_std_eta=n_max_std_eta,
        device=device,
        dtype=dtype,
        generator=generator,
    )

    return transform([axis1, axis2], [angle1, angle2])


def rand_rotation(
    shape: List[int],
    device: str = "cpu",
    dtype: torch.dtype = torch.float32,
    generator: torch.Generator = None,
):
    """
    Create rotation matrices embedded in Lorentz transformations.
    The rotations are sampled uniformly using quaternions.

    Parameters
    ----------
    shape: List[int]
        Shape of the transformation matrices
    device: str
    dtype: torch.dtype
    generator: torch.Generator

    Returns
    -------
    final_trafo: torch.tensor
        The resulting Lorentz transformation matrices of shape (*shape, 4, 4).
    """
    # generate random quaternions
    u = torch.rand(*shape, 3, device=device, dtype=dtype, generator=generator)
    q1 = torch.sqrt(1 - u[..., 0]) * torch.sin(2 * torch.pi * u[..., 1])
    q2 = torch.sqrt(1 - u[..., 0]) * torch.cos(2 * torch.pi * u[..., 1])
    q3 = torch.sqrt(u[..., 0]) * torch.sin(2 * torch.pi * u[..., 2])
    q0 = torch.sqrt(u[..., 0]) * torch.cos(2 * torch.pi * u[..., 2])

    # create rotation matrix from quaternions
    # see https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation
    R1 = torch.stack(
        [1 - 2 * (q2**2 + q3**2), 2 * (q1 * q2 - q0 * q3), 2 * (q1 * q3 + q0 * q2)],
        dim=-1,
    )
    R2 = torch.stack(
        [2 * (q1 * q2 + q0 * q3), 1 - 2 * (q1**2 + q3**2), 2 * (q2 * q3 - q0 * q1)],
        dim=-1,
    )
    R3 = torch.stack(
        [2 * (q1 * q3 - q0 * q2), 2 * (q2 * q3 + q0 * q1), 1 - 2 * (q1**2 + q2**2)],
        dim=-1,
    )
    R = torch.stack([R1, R2, R3], dim=-2)

    trafo = torch.eye(4, device=device, dtype=dtype).expand(*shape, 4, 4).clone()
    trafo[..., 1:, 1:] = R
    return trafo


def rand_boost(
    shape: List[int],
    std_eta: float = 0.1,
    n_max_std_eta: float = 3.0,
    device: str = "cpu",
    dtype: torch.dtype = torch.float32,
    generator: torch.Generator = None,
):
    """
    Create a general pure boost (symmetric Lorentz transformation).

    Parameters
    ----------
    shape: List[int]
        Shape of the transformation matrices
    std_eta: float
        Standard deviation of rapidity
    n_max_std_eta: float
        Allowed number of standard deviations;
        used to sample from a truncated Gaussian
    device: str
    dtype: torch.dtype
    generator: torch.Generator

    Returns
    -------
    final_trafo: torch.tensor
        The resulting Lorentz transformation matrices of shape (*shape, 4, 4).
    """
    shape = list(shape) + [3]
    beta = sample_rapidity(
        shape,
        std_eta,
        n_max_std_eta,
        device=device,
        dtype=dtype,
        generator=generator,
    )
    beta2 = (beta**2).sum(dim=-1, keepdim=True)
    gamma = 1 / (1 - beta2).clamp(min=1e-10).sqrt()
    fourmomenta = torch.cat([gamma, beta], axis=-1)

    boost = restframe_boost(fourmomenta)
    return boost


def sample_rapidity(
    shape,
    std_eta,
    n_max_std_eta=3.0,
    device="cpu",
    dtype=torch.float32,
    generator: torch.Generator = None,
):
    """Sample rapidity from a gaussian distribution.

    Parameters
    ----------
    shape: List[int]
        Shape of the output tensor
    std_eta: float
        Standard deviation of the rapidity
    n_max_std_eta: float
        Maximum number of standard deviations for truncation
    device: str
    dtype: torch.dtype
    generator: torch.Generator
    """
    angle = (
        torch.randn(*shape, device=device, dtype=dtype, generator=generator) * std_eta
    )
    truncate_mask = torch.abs(angle) > std_eta * n_max_std_eta
    while truncate_mask.any():
        new_angle = (
            torch.randn(*shape, device=device, dtype=dtype, generator=generator)
            * std_eta
        )
        angle[truncate_mask] = new_angle[truncate_mask]
        truncate_mask = torch.abs(angle) > std_eta * n_max_std_eta
    return angle



import torch
import pytest
from tests.constants import TOLERANCES, BATCH_DIMS

from lloca.utils.lorentz import lorentz_inner, lorentz_squarednorm
from lloca.utils.orthogonalize_4d import orthogonalize_4d


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("method", ["cross", "gramschmidt"])
@pytest.mark.parametrize(
    "vector_type,eps",
    [
        ("naive", None),
        ("coplanar", 1e-2),
        ("coplanar", 1e-3),
        ("lightlike", 1e-2),
        ("lightlike", 1e-4),
    ],
)
def test_orthogonalize(batch_dims, method, vector_type, eps):
    # check orthogonality after using the function
    dtype = torch.float64

    v1 = torch.randn(batch_dims + [4], dtype=dtype)
    v2 = torch.randn(batch_dims + [4], dtype=dtype)
    if vector_type == "naive":
        # third vector is fully random
        v3 = torch.randn(batch_dims + [4], dtype=dtype)
    elif vector_type == "coplanar":
        # third vector is almost a linear combination of the first 2 vectors
        c1, c2 = torch.randn(batch_dims, dtype=dtype), torch.randn(
            batch_dims, dtype=dtype
        )
        v3 = (
            v1
            + c1.unsqueeze(-1) * v1
            + c2.unsqueeze(-1) * v2
            + eps * torch.randn_like(v1)
        )
    elif vector_type == "lightlike":
        # make all vectors lightlike
        v3s = torch.randn([3] + batch_dims + [3], dtype=dtype)
        norm = eps * torch.randn_like(v3s[..., [0]])
        v0 = torch.sqrt((v3s**2).sum(dim=-1, keepdim=True)) + norm
        vecs = torch.cat([v0, v3s], dim=-1)
        v1, v2, v3 = vecs

    orthogonal_vecs = orthogonalize_4d([v1, v2, v3], method=method)

    # test orthogonality
    n = orthogonal_vecs.ndim
    perm = [n - 2] + list(range(0, n - 2)) + [n - 1]
    orthogonal_vecs = orthogonal_vecs.permute(*perm)
    for i1, v1 in enumerate(orthogonal_vecs):
        for i2, v2 in enumerate(orthogonal_vecs):
            inner = lorentz_inner(v1, v2)
            target = torch.ones_like(inner) if i1 == i2 else torch.zeros_like(inner)
            torch.testing.assert_close(inner.abs(), target, **TOLERANCES)

    # check that there is only one time-like vector in each triplet of orthogonalized vectors
    norm = torch.stack([lorentz_squarednorm(v) for v in orthogonal_vecs], dim=-1)
    num_timelike = torch.sum(norm > 0, dim=-1)
    torch.testing.assert_close(
        num_timelike, torch.ones_like(num_timelike), atol=0, rtol=0
    )



from typing import Any, Dict
import torch
from torch_geometric.nn import MessagePassing

from ..framesnet.frames import ChangeOfFrames, Frames, IndexSelectFrames
from ..reps.tensorreps_transform import TensorRepsTransform


class LLoCaMessagePassing(MessagePassing):
    """Adaptation of the torch_geometric MessagePassing class using the LLoCa formalism."""

    def __init__(self, params_dict: Dict[str, Dict[str, Any]], aggr="add") -> None:
        """Initializes a new instance of the LLoCaMessagePassing class.

        Parameters
        ----------
        params_dict: Dict[str, Dict[str, Any]]
            A dictionary containing the parameters for the message passing algorithm and the corresponding representations.
            Each key in the dictionary represents a feature, and the value is another dictionary with keys "type" and "rep".
            The "type" can be either "local" or "global", and "rep" is an instance of TensorRepsTransform that defines how to transform the features.
        aggr: str, optional
            The aggregation method to use for combining messages. Defaults to "add".
        """
        super().__init__(aggr=aggr)

        self.params_dict = params_dict

        tmp_dict = {}

        for key, value in self.params_dict.items():
            if value["type"] is not None:
                tmp_dict[key] = TensorRepsTransform(value["rep"])

        self.transform_dict = torch.nn.ModuleDict(tmp_dict)

        # Register hooks to call before propagating and before sending messages
        self.register_propagate_forward_pre_hook(self.pre_propagate_hook)
        self.register_message_forward_pre_hook(self.pre_message_hook)

    def pre_propagate_hook(self, module: Any, inputs: tuple) -> tuple:
        """A hook method called before propagating messages in the message passing algorithm. We
        save the frames in the class variable and remove it from the inputs dictionary.
        """
        assert (
            inputs[-1].get("frames") is not None
        ), "frames are not in the propagate inputs"

        self._frames = inputs[-1]["frames"]
        self._edge_index = inputs[0]

        return inputs

    def pre_message_hook(self, module: Any, inputs: tuple) -> tuple:
        """Pre-message hook method that is called before passing messages in the message passing
        algorithm. We transform the features according to the representations in the params_dict.
        """

        # calculate frames_i, frames_j and the U matrix
        if isinstance(self._frames, tuple):
            frames_i = IndexSelectFrames(self._frames[1], self._edge_index[1])
            frames_j = IndexSelectFrames(self._frames[0], self._edge_index[0])
        elif isinstance(self._frames, Frames):
            frames_i = IndexSelectFrames(self._frames, self._edge_index[1])
            frames_j = IndexSelectFrames(self._frames, self._edge_index[0])
        else:
            raise ValueError(
                f"frames should be either a tuple or an Frames object but is {type(self._frames)}"
            )

        U = ChangeOfFrames(frames_start=frames_j, frames_end=frames_i)

        # now go through the params_dict and get the representations and transform the features in the right way
        for param, param_info in self.params_dict.items():
            if param_info["type"] == "local":
                assert param + "_j" in inputs[-1], f"Key {param}_j not in inputs"
                # transform the features according to the representation
                inputs[-1][param + "_j"] = self.transform_dict[param](
                    inputs[-1][param + "_j"], U
                )
            elif param_info["type"] == "global":
                if inputs[-1].get(param) is not None:
                    inputs[-1][param] = self.transform_dict[param](
                        inputs[-1][param], frames_i
                    )
                if inputs[-1].get(param + "_j") is not None:
                    inputs[-1][param + "_j"] = self.transform_dict[param](
                        inputs[-1][param + "_j"], frames_i
                    )
                if inputs[-1].get(param + "_i") is not None:
                    inputs[-1][param + "_i"] = self.transform_dict[param](
                        inputs[-1][param + "_i"], frames_i
                    )

        return inputs



# Equivariant local frames, constructed in various ways
import torch
from torch_geometric.utils import scatter

from .frames import Frames
from .nonequi_frames import FramesPredictor
from ..utils.utils import get_batch_from_ptr
from ..utils.polar_decomposition import polar_decomposition
from ..utils.lorentz import lorentz_eye, lorentz_squarednorm
from ..utils.orthogonalize_4d import orthogonalize_4d


class LearnedFrames(FramesPredictor):
    """Abstract class for local Frames constructed
    based on equivariantly predicted vectors"""

    def __init__(
        self,
        equivectors,
        n_vectors,
        is_global=False,
        random=False,
        fix_params=False,
        ortho_kwargs={},
    ):
        """
        Parameters
        ----------
        equivectors: nn.Module
            Network that equivariantly predicts vectors
        n_vectors: int
            Number of vectors to predict
        is_global: bool
            If True, average the predicted vectors to construct a global frame
        random: bool
            If True, re-initialize the equivectors at each forward pass
            This is a fancy way of doing data augmentation
        fix_params: bool
            Like random, but without the resampling
        ortho_kwargs: dict
            Keyword arguments for orthogonalization
        """
        super().__init__()
        self.ortho_kwargs = ortho_kwargs
        self.equivectors = equivectors(n_vectors=n_vectors)
        self.is_global = is_global
        self.random = random
        if random or fix_params:
            self.equivectors.requires_grad_(False)

    def init_weights_or_not(self):
        if self.random and self.training:
            self.equivectors.apply(init_weights)

    def globalize_vecs_or_not(self, vecs, ptr):
        return average_event(vecs, ptr) if self.is_global else vecs

    def __repr__(self):
        classname = self.__class__.__name__
        method = self.ortho_kwargs["method"]
        string = f"{classname}(method={method})"
        return string


class LearnedPDFrames(LearnedFrames):
    """Construct Frames as learnable polar decomposition (boost+rotation).
    This is our default approach.
    LearnedSO13Frames works similarly well, but is less flexible.
    """

    def __init__(
        self,
        *args,
        gamma_max=None,
        gamma_hardness=None,
        deterministic_boost=None,
        **kwargs,
    ):
        super().__init__(*args, n_vectors=3, **kwargs)
        self.gamma_max = gamma_max
        self.gamma_hardness = gamma_hardness
        self.deterministic_boost = deterministic_boost

    def forward(self, fourmomenta, scalars=None, ptr=None, return_tracker=False):
        """
        Parameters
        ----------
        fourmomenta: torch.Tensor
            Tensor of shape (..., 4) containing the four-momenta
        scalars: torch.Tensor or None
            Optional tensor of shape (..., n_scalars) containing additional scalar features
        ptr: torch.Tensor or None
            Pointer for sparse tensors, or None for dense tensors
        return_tracker: bool
            If True, return a tracker dictionary with regularization information

        Returns
        -------
        Frames
            Local frames constructed from the polar decomposition of the four-momenta
        tracker: dict (optional)
            Dictionary containing regularization information, if return_tracker is True
        """
        self.init_weights_or_not()
        vecs = self.equivectors(fourmomenta, scalars=scalars, ptr=ptr)
        vecs = self.globalize_vecs_or_not(vecs, ptr)
        boost = vecs[..., 0, :]
        rotation_references = [vecs[..., i, :] for i in range(1, vecs.shape[-2])]
        boost = self._deterministic_boost(boost, ptr)
        boost, reg_gammamax, gamma_mean, gamma_max = self._clamp_boost(boost)

        trafo, reg_collinear = polar_decomposition(
            boost,
            rotation_references,
            **self.ortho_kwargs,
            return_reg=True,
        )
        tracker = {
            "reg_collinear": reg_collinear,
            "gamma_mean": gamma_mean,
            "gamma_max": gamma_max,
        }
        if reg_gammamax is not None:
            tracker["reg_gammamax"] = reg_gammamax
        frames = Frames(trafo, is_global=self.is_global)
        return (frames, tracker) if return_tracker else frames

    def _clamp_boost(self, x):
        mass = lorentz_squarednorm(x).clamp(min=0).sqrt().unsqueeze(-1)
        beta = x[..., 1:] / x[..., [0]].clamp(min=1e-10)
        gamma = x[..., [0]] / mass
        gamma_max = gamma.max().detach().cpu()
        gamma_mean = gamma.detach().mean().cpu()

        if self.gamma_max is None:
            return x, None, gamma_mean, gamma_max

        else:
            # carefully clamp gamma to keep boosts under control
            reg_gammamax = (gamma > self.gamma_max).sum().cpu()
            gamma_reg = soft_clamp(
                gamma, min=1, max=self.gamma_max, hardness=self.gamma_hardness
            )
            beta_scaling = (
                torch.sqrt(
                    torch.clamp(1 - 1 / gamma_reg.clamp(min=1e-10).square(), min=1e-10)
                )
                / (beta**2).sum(dim=-1, keepdim=True).clamp(min=1e-10).sqrt()
            )
            beta_reg = beta * beta_scaling
            x_reg = mass * torch.cat((gamma_reg, gamma_reg * beta_reg), dim=-1)
            return x_reg, reg_gammamax, gamma_mean, gamma_max

    def _deterministic_boost(self, boost, ptr):
        if self.deterministic_boost is None:
            pass
        elif self.deterministic_boost == "global":
            # average boost vector over the event
            boost = average_event(boost, ptr)
        elif self.deterministic_boost == "local":
            # average boost over all other particles in the event
            boost_averaged = average_event(boost, ptr)
            if ptr is None:
                nparticles = boost.shape[1]
            else:
                diff = ptr[1:] - ptr[:-1]
                nparticles = (diff).repeat_interleave(diff).unsqueeze(-1)
            boost = boost_averaged - boost / nparticles
        else:
            raise ValueError(
                f"Option deterministic_boost={self.deterministic_boost} not implemented"
            )

        return boost


class LearnedSO13Frames(LearnedFrames):
    """
    Local frames from an orthonormal set of Lorentz vectors
    constructed from equivariantly predicted vectors
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, n_vectors=3, **kwargs)

    def forward(self, fourmomenta, scalars=None, ptr=None, return_tracker=False):
        """
        Parameters
        ----------
        fourmomenta: torch.Tensor
            Tensor of shape (..., 4) containing the four-momenta
        scalars: torch.Tensor or None
            Optional tensor of shape (..., n_scalars) containing additional scalar features
        ptr: torch.Tensor or None
            Pointer for sparse tensors, or None for dense tensors
        return_tracker: bool
            If True, return a tracker dictionary with regularization information

        Returns
        -------
        Frames
            Local frames constructed from the polar decomposition of the four-momenta
        tracker: dict (optional)
            Dictionary containing regularization information, if return_tracker is True
        """
        self.init_weights_or_not()
        vecs = self.equivectors(fourmomenta, scalars=scalars, ptr=ptr)
        vecs = self.globalize_vecs_or_not(vecs, ptr)
        vecs = [vecs[..., i, :] for i in range(vecs.shape[-2])]

        trafo, reg_lightlike, reg_coplanar = orthogonalize_4d(
            vecs, **self.ortho_kwargs, return_reg=True
        )

        tracker = {"reg_lightlike": reg_lightlike, "reg_coplanar": reg_coplanar}
        frames = Frames(trafo, is_global=self.is_global)
        return (frames, tracker) if return_tracker else frames


class LearnedRestFrames(LearnedFrames):
    """Rest frame transformation with learnable equivariant rotation.
    This is a special case of LearnedPolarDecompositionFrames
    where the boost vector is chosen to be the particle momentum."""

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, n_vectors=2, **kwargs)

    def forward(self, fourmomenta, scalars=None, ptr=None, return_tracker=False):
        """
        Parameters
        ----------
        fourmomenta: torch.Tensor
            Tensor of shape (..., 4) containing the four-momenta
        scalars: torch.Tensor or None
            Optional tensor of shape (..., n_scalars) containing additional scalar features
        ptr: torch.Tensor or None
            Pointer for sparse tensors, or None for dense tensors
        return_tracker: bool
            If True, return a tracker dictionary with regularization information

        Returns
        -------
        Frames
            Local frames constructed from the polar decomposition of the four-momenta
        tracker: dict (optional)
            Dictionary containing regularization information, if return_tracker is True
        """
        self.init_weights_or_not()
        references = self.equivectors(fourmomenta, scalars=scalars, ptr=ptr)
        references = self.globalize_vecs_or_not(references, ptr)
        references = [references[..., i, :] for i in range(references.shape[-2])]

        trafo, reg_collinear = polar_decomposition(
            fourmomenta,
            references,
            **self.ortho_kwargs,
            return_reg=True,
        )
        tracker = {"reg_collinear": reg_collinear}
        frames = Frames(trafo, is_global=self.is_global)
        return (frames, tracker) if return_tracker else frames


class LearnedSO3Frames(LearnedFrames):
    """Local frames under rotations for SO(3)-equivariant architectures.
    This is a special case of LearnedPolarDecompositionFrames
    where the first vector is trivial (1,0,0,0)."""

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        self.n_vectors = 2
        super().__init__(
            *args,
            n_vectors=self.n_vectors,
            **kwargs,
        )

    def forward(self, fourmomenta, scalars=None, ptr=None, return_tracker=False):
        """
        Parameters
        ----------
        fourmomenta: torch.Tensor
            Tensor of shape (..., 4) containing the four-momenta
        scalars: torch.Tensor or None
            Optional tensor of shape (..., n_scalars) containing additional scalar features
        ptr: torch.Tensor or None
            Pointer for sparse tensors, or None for dense tensors
        return_tracker: bool
            If True, return a tracker dictionary with regularization information

        Returns
        -------
        Frames
            Local frames constructed from the polar decomposition of the four-momenta
        tracker: dict (optional)
            Dictionary containing regularization information, if return_tracker is True
        """
        self.init_weights_or_not()
        references = self.equivectors(fourmomenta, scalars=scalars, ptr=ptr)
        references = self.globalize_vecs_or_not(references, ptr)
        fourmomenta = lorentz_eye(
            fourmomenta.shape[:-1], device=fourmomenta.device, dtype=fourmomenta.dtype
        )[
            ..., 0
        ]  # only difference compared to LearnedPolarDecompositionFrames
        references = [references[..., i, :] for i in range(self.n_vectors)]

        trafo, reg_collinear = polar_decomposition(
            fourmomenta,
            references,
            **self.ortho_kwargs,
            return_reg=True,
        )
        tracker = {"reg_collinear": reg_collinear}
        frames = Frames(trafo, is_global=self.is_global)
        return (frames, tracker) if return_tracker else frames


class LearnedSO2Frames(LearnedFrames):
    """Local frames for SO(2)-equivariant architectures
    (equivariant under rotations around the beam axis).
    This is a special case of LearnedPolarDecompositionFrames
    where the firsts two vectors are trivial (1,0,0,0) and (0,0,0,1)."""

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        self.n_vectors = 1
        super().__init__(
            *args,
            n_vectors=self.n_vectors,
            **kwargs,
        )

    def forward(self, fourmomenta, scalars=None, ptr=None, return_tracker=False):
        """
        Parameters
        ----------
        fourmomenta: torch.Tensor
            Tensor of shape (..., 4) containing the four-momenta
        scalars: torch.Tensor or None
            Optional tensor of shape (..., n_scalars) containing additional scalar features
        ptr: torch.Tensor or None
            Pointer for sparse tensors, or None for dense tensors
        return_tracker: bool
            If True, return a tracker dictionary with regularization information

        Returns
        -------
        Frames
            Local frames constructed from the polar decomposition of the four-momenta
        tracker: dict (optional)
            Dictionary containing regularization information, if return_tracker is True
        """
        self.init_weights_or_not()
        references = self.equivectors(fourmomenta, scalars=scalars, ptr=ptr)
        extra_references = self.globalize_vecs_or_not(references, ptr)
        fourmomenta = lorentz_eye(
            fourmomenta.shape[:-1], device=fourmomenta.device, dtype=fourmomenta.dtype
        )[
            ..., 0
        ]  # difference 1 compared LearnedPolarDecompositionFrames
        references = [
            lorentz_eye(
                fourmomenta.shape[:-1],
                device=fourmomenta.device,
                dtype=fourmomenta.dtype,
            )[..., 3]
        ]  # difference 2 compared LearnedPolarDecompositionFrames
        references.append(extra_references[..., 0, :])

        trafo, reg_collinear = polar_decomposition(
            fourmomenta,
            references,
            **self.ortho_kwargs,
            return_reg=True,
        )

        tracker = {"reg_collinear": reg_collinear}
        frames = Frames(trafo, is_global=self.is_global)
        return (frames, tracker) if return_tracker else frames


def average_event(vecs, ptr=None):
    """Average vectors across events and expand again.

    Parameters
    ----------
    vecs: torch.Tensor
        Tensor of shape (..., n_vectors, 4)
        where the last dimension contains the vectors
    ptr: torch.Tensor or None
        Pointer to the batch of events, or None for global averaging

    Returns
    -------
    torch.Tensor
        Averaged vectors of shape (..., n_vectors, 4).
    """
    if ptr is None:
        vecs = vecs.mean(dim=1, keepdim=True).expand_as(vecs)
    else:
        batch = get_batch_from_ptr(ptr)
        vecs = scatter(vecs, batch, dim=0, reduce="mean").index_select(0, batch)
    return vecs


def init_weights(module):
    if hasattr(module, "reset_parameters"):
        module.reset_parameters()


def soft_clamp(x, max=None, min=None, hardness=None):
    if hardness is None:
        # hard clamp
        return x.clamp(min=min, max=max)
    else:
        # soft clamp (better gradients)
        out = max - torch.nn.functional.softplus(max - x, beta=hardness)
        return out.clamp(min=min)



# Bookkeeping class for convenient access to local frames
import torch

from ..utils.lorentz import lorentz_eye, lorentz_metric


class Frames:
    """
    Collection of Lorentz transformations, represented as (*dims, 4, 4) matrices.
    Expensive properties like det and inv are cached for performance.
    Attributes should not be changed after initialization to avoid inconsistencies.
    Shapes can be modified with e.g. .reshape(), .expand() and .repeat().
    """

    def __init__(
        self,
        matrices: torch.Tensor = None,
        is_global: bool = False,
        det: torch.Tensor = None,
        inv: torch.Tensor = None,
        is_identity: bool = False,
        shape=None,
        device: str = None,
        dtype: torch.dtype = None,
    ):
        """There are 2 ways to initialize an Frames object:
        - From matrices: Set matrices and optionally is_global, det, inv
        - As identity: Set is_identity=True, shape, device, dtype

        Parameters
        ----------
        matrices: torch.tensor
            Transformation matrices of shape (*dims, 4, 4)
        is_global: bool
            Whether frames are the same for all particles in the point cloud
        inv: torch.Tensor
            Optional cached inverse of shape (*dims, 4, 4).
            If not given, takes a bit of extra time to compute.
        det: torch.Tensor
            Optional cached determinant of shape (*dims).
            If not given, takes a bit of extra time to compute.
        is_identity: bool
            Sets matrices to diagonal.
        shape: List[int]
            Specifies matrices.shape[:-2] if is_identity. Otherwise inferred from matrices.
        device: str
            Specifies device if is_identity. Otherwise inferred from matrices.
        dtype: torch.dtype
            Specifies dtype if is_identity. Otherwise inferred from matrices.
        """
        # straight-forward initialization
        self.is_identity = is_identity
        if is_identity:
            if matrices is None:
                assert shape and device and dtype
            else:
                shape = matrices.shape[:-2]
                device = matrices.device
                dtype = matrices.dtype

            self.matrices = lorentz_eye(shape, device=device, dtype=dtype)
            self.is_global = True
            self.det = torch.ones(self.shape[:-2], dtype=self.dtype, device=self.device)
            self.inv = self.matrices
        else:
            assert matrices is not None
            assert matrices.shape[-2:] == (
                4,
                4,
            ), f"matrices must be of shape (..., 4, 4), but found {matrices.shape[-2:]} instead"

            self.matrices = matrices
            self.is_global = is_global
            if det is not None:
                assert det.shape == matrices.shape[:-2]
            if inv is not None:
                assert inv.shape == matrices.shape
            self.det = det
            self.inv = inv

        # cache expensive properties
        self.metric = lorentz_metric(
            self.shape[:-2], device=self.device, dtype=self.dtype
        )
        if self.det is None:
            self.det = torch.linalg.det(self.matrices)
        if self.inv is None:
            self.inv = self.metric @ self.matrices.transpose(-1, -2) @ self.metric

    def __repr__(self):
        return repr(self.matrices)

    def reshape(self, *shape):
        """Reshape the matrices to a new shape.

        Parameters
        ----------
        shape: int
            New shape for the matrices. The last two dimensions must be (4, 4).

        Returns
        -------
        Frames
        """
        assert shape[-2:] == (4, 4)
        return Frames(
            matrices=self.matrices.reshape(*shape),
            is_identity=self.is_identity,
            is_global=self.is_global,
            inv=self.inv.reshape(*shape),
            det=self.det.reshape(*shape[:-2]),
        )

    def expand(self, *shape):
        """Expand the matrices to a new shape.

        Parameters
        ----------
        shape: int
            New shape for the matrices. The last two dimensions must be (4, 4).

        Returns
        -------
        Frames
        """
        assert shape[-2:] == (4, 4)
        return Frames(
            matrices=self.matrices.expand(*shape),
            is_identity=self.is_identity,
            is_global=self.is_global,
            inv=self.inv.expand(*shape),
            det=self.det.expand(*shape[:-2]),
        )

    def repeat(self, *shape):
        """Repeat the matrices to a new shape.

        Parameters
        ----------
        shape: int
            New shape for the matrices. The last two dimensions must be (4, 4).

        Returns
        -------
        Frames
        """
        assert shape[-2:] == (1, 1)
        return Frames(
            matrices=self.matrices.repeat(*shape),
            is_identity=self.is_identity,
            is_global=self.is_global,
            inv=self.inv.repeat(*shape),
            det=self.det.repeat(*shape[:-2]),
        )

    def to(self, dtype=None, device=None):
        self.matrices = self.matrices.to(device=device, dtype=dtype)
        self.inv = self.inv.to(device=device, dtype=dtype)
        self.det = self.det.to(device=device, dtype=dtype)

    @property
    def device(self):
        return self.matrices.device

    @property
    def dtype(self):
        return self.matrices.dtype

    @property
    def shape(self):
        return self.matrices.shape


class InverseFrames(Frames):
    """Inverse of a collection of transformations."""

    def __init__(self, frames: Frames):
        super().__init__(
            matrices=frames.inv,
            is_global=frames.is_global,
            inv=frames.matrices,
            det=frames.det,
            is_identity=frames.is_identity,
            device=frames.device,
            dtype=frames.dtype,
            shape=frames.shape,
        )


class IndexSelectFrames(Frames):
    """Index-controlled selection of transformation matrices from an Frames object."""

    def __init__(self, frames: Frames, indices: torch.Tensor):
        super().__init__(
            matrices=frames.matrices.index_select(0, indices),
            is_global=frames.is_global,
            inv=frames.inv.index_select(0, indices),
            det=frames.det.index_select(0, indices),
            is_identity=frames.is_identity,
            device=frames.device,
            dtype=frames.dtype,
            shape=frames.shape,
        )


class ChangeOfFrames(Frames):
    """
    Change of frames between two Frames objects
    Formally, for L_start and L_end we have
    L_change = L_end * L_start^-1

    WARNING: This function does not mix the frames of different point clouds
    It is used in TFMessagePassing where this mixing is performed using the edge_index
    """

    def __init__(self, frames_start: Frames, frames_end: Frames):
        assert frames_start.shape == frames_end.shape
        if frames_start.is_global:
            super().__init__(
                is_identity=True,
                shape=frames_start.shape,
                device=frames_start.device,
                dtype=frames_start.dtype,
            )
        else:
            super().__init__(
                matrices=frames_end.matrices @ frames_start.inv,
                is_global=False,
                inv=frames_start.matrices @ frames_end.inv,
                det=frames_start.det * frames_end.det,
            )


class LowerIndicesFrames(Frames):
    """
    Frames with lower indices
    Used in LLoCaAttention to lower the key indices
    """

    def __init__(self, frames):
        super().__init__(
            matrices=frames.metric @ frames.matrices,
            inv=frames.inv @ frames.metric,
            det=-frames.det,
            is_global=frames.is_global,
            is_identity=frames.is_identity,
            device=frames.device,
            dtype=frames.dtype,
            shape=frames.shape,
        )



import torch
import pytest
from tests.constants import TOLERANCES, BATCH_DIMS, LOGM2_MEAN_STD
from tests.helpers import sample_particle, lorentz_test

from lloca.utils.polar_decomposition import (
    restframe_boost,
    polar_decomposition,
)
from lloca.utils.lorentz import lorentz_squarednorm
from lloca.utils.rand_transforms import rand_lorentz, rand_rotation


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize(
    "restframe_transform",
    [restframe_boost, polar_decomposition],
)
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_restframe(batch_dims, restframe_transform, logm2_std, logm2_mean):
    dtype = torch.float64

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)
    kwargs = {}
    if restframe_transform == polar_decomposition:
        kwargs["references"] = [
            sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)
            for _ in range(2)
        ]
        kwargs["return_reg"] = False

    # determine transformation into rest frame
    rest_trafo = restframe_transform(fm, **kwargs)
    fm_rest = torch.einsum("...ij,...j->...i", rest_trafo, fm)

    # check that the transformed fourmomenta are in the rest frame,
    # i.e. their spatial components vanish and the temporal component is the mass
    torch.testing.assert_close(
        fm_rest[..., 1:], torch.zeros_like(fm[..., 1:]), **TOLERANCES
    )
    torch.testing.assert_close(
        fm_rest[..., 0] ** 2, lorentz_squarednorm(fm), **TOLERANCES
    )

    lorentz_test(rest_trafo, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("random_transform", [rand_lorentz, rand_rotation])
@pytest.mark.parametrize("logm2_mean,logm2_std", LOGM2_MEAN_STD)
def test_restframe_transformation(batch_dims, random_transform, logm2_std, logm2_mean):
    dtype = torch.float64

    # sample Lorentz vectors
    fm = sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)
    references = [
        sample_particle(batch_dims, logm2_std, logm2_mean, dtype=dtype)
        for _ in range(2)
    ]

    # determine transformation into rest frame
    rest_trafo = polar_decomposition(fm, references, return_reg=False)
    fm_rest = torch.einsum("...ij,...j->...i", rest_trafo, fm)

    # random global transformation
    random = random_transform([1], dtype=dtype)
    random = random.repeat(*batch_dims, 1, 1)
    fm_prime = torch.einsum("...ij,...j->...i", random, fm)
    references_prime = [torch.einsum("...ij,...j->...i", random, r) for r in references]
    rest_trafo_prime = polar_decomposition(fm_prime, references_prime, return_reg=False)
    fm_rest_prime = torch.einsum("...ij,...j->...i", rest_trafo_prime, fm_prime)

    # check that fourmomenta in rest frame is invariant
    torch.testing.assert_close(fm_rest, fm_rest_prime, **TOLERANCES)

    # check that the restframe transformations transform as expected
    # expect rest_trafo_prime = rest_trafo * random^-1
    # or rest_trafo_prime * random = rest_trafo
    rest_trafo_expected = torch.einsum("...ij,...jk->...ik", rest_trafo_prime, random)

    torch.testing.assert_close(rest_trafo, rest_trafo_expected, **TOLERANCES)



import torch

from lloca.utils.lorentz import lorentz_metric
from lloca.equivectors.equimlp import EquiMLP


def sample_particle(
    shape, logm2_std, logm2_mean, device=torch.device("cpu"), dtype=torch.float32
):
    assert logm2_std > 0
    logm2 = torch.randn(*shape, 1, device=device, dtype=dtype) * logm2_std + logm2_mean
    p3 = torch.randn(*shape, 3, device=device, dtype=dtype)
    E = torch.sqrt(logm2.exp() + (p3**2).sum(dim=-1, keepdim=True))
    return torch.cat([E, p3], dim=-1)


def lorentz_test(trafo, **kwargs):
    """
    Test that the transformation matrix T is orthogonal

    Condition: T^T * g * T = g
    with the Lorentz metric g = diag(1, -1, -1, -1)
    """
    metric = lorentz_metric(trafo.shape[:-2], trafo.device, trafo.dtype)
    test = torch.einsum(
        "...ij,...jk,...kl->...il", trafo, metric, trafo.transpose(-1, -2)
    )
    torch.testing.assert_close(test, metric, **kwargs)


def equivectors_builder(num_scalars=0):
    def builder(n_vectors):
        return EquiMLP(
            n_vectors=n_vectors,
            num_scalars=num_scalars,
            hidden_channels=16,
            num_layers_mlp=1,
            num_blocks=1,
        )

    return builder



import torch
from torch import nn
import math
from torch_geometric.nn import MessagePassing
from torch_geometric.utils import scatter, softmax

from .base import EquiVectors
from ..backbone.mlp import MLP
from ..utils.lorentz import lorentz_squarednorm
from ..utils.utils import (
    build_edge_index_fully_connected,
    get_edge_index_from_ptr,
    get_edge_attr,
    get_ptr_from_batch,
)


class EquiEdgeConv(MessagePassing):
    def __init__(
        self,
        in_vectors,
        out_vectors,
        num_scalars,
        hidden_channels,
        num_layers_mlp,
        include_edges=True,
        operation="add",
        nonlinearity="exp",
        fm_norm=False,
        dropout_prob=None,
        aggr="sum",
        layer_norm=False,
    ):
        """Equivariant edge convolution, implemented using PyTorch Geometric's MessagePassing class.

        Parameters
        ----------
        in_vectors : int
            Number of input vectors.
        out_vectors : int
            Number of output vectors.
        num_scalars : int
            Number of scalar features per particle.
        hidden_channels : int
            Number of hidden channels in the MLP.
        num_layers_mlp : int
            Number of hidden layers in the MLP.
        include_edges : bool, optional
            Whether to include edge attributes in the message passing. If True, edge attributes will be calculated from fourmomenta and standardized. Default is True.
        operation : str, optional
            Operation to perform on the fourmomenta. Options are "add", "diff", or "single". Default is "add".
        nonlinearity : str, optional
            Nonlinearity to apply to the output of the MLP. Options are None, "exp", "softplus" and "softmax".
        """
        super().__init__(aggr=aggr, flow="target_to_source")
        assert num_scalars > 0 or include_edges
        self.include_edges = include_edges
        self.layer_norm = layer_norm
        self.operation = self.get_operation(operation)
        self.nonlinearity = self.get_nonlinearity(nonlinearity)
        self.fm_norm = fm_norm
        assert not (operation == "single" and fm_norm)  # unstable

        in_edges = in_vectors if include_edges else 0
        in_channels = 2 * num_scalars + in_edges
        self.mlp = MLP(
            in_shape=[in_channels],
            out_shape=out_vectors,
            hidden_channels=hidden_channels,
            hidden_layers=num_layers_mlp,
            dropout_prob=dropout_prob,
        )

        if include_edges:
            self.register_buffer("edge_inited", torch.tensor(False, dtype=torch.bool))
            self.register_buffer("edge_mean", torch.tensor(0.0))
            self.register_buffer("edge_std", torch.tensor(1.0))

    def forward(self, fourmomenta, scalars, edge_index, batch=None):
        """
        Parameters
        ----------
        fourmomenta : torch.Tensor
            Tensor of shape (num_particles, in_vectors*4) containing the fourmomenta of the particles.
        scalars : torch.Tensor
            Tensor of shape (num_particles, num_scalars) containing scalar features for each particle.
        edge_index : torch.Tensor
            Edge index tensor containing the indices of the source and target nodes, shape (2, num_edges).
        batch : torch.Tensor, optional
            Batch tensor indicating the batch each particle belongs to. If None, all particles are assumed to belong to the same batch.

        Returns
        -------
        torch.Tensor
            Tensor of shape (num_particles, out_vectors*4) containing the predicted vectors for each edge.
        """
        # calculate and standardize edge attributes
        fourmomenta = fourmomenta.reshape(scalars.shape[0], -1, 4)
        if self.include_edges:
            edge_attr = get_edge_attr(fourmomenta, edge_index)
            if not self.edge_inited:
                self.edge_mean = edge_attr.mean().detach()
                self.edge_std = edge_attr.std().clamp(min=1e-5).detach()
                self.edge_inited.fill_(True)
            edge_attr = (edge_attr - self.edge_mean) / self.edge_std
            edge_attr = edge_attr.reshape(edge_attr.shape[0], -1)

            # related to fourmomenta_float64 option
            edge_attr = edge_attr.to(scalars.dtype)
        else:
            edge_attr = None

        fourmomenta = fourmomenta[:, 0, :]
        # message-passing
        vecs = self.propagate(
            edge_index, s=scalars, fm=fourmomenta, edge_attr=edge_attr, batch=batch
        )
        # equivariant layer normalization
        if self.layer_norm:
            norm = lorentz_squarednorm(vecs.reshape(fourmomenta.shape[0], -1, 4))
            norm = norm.sum(dim=-1).unsqueeze(-1)
            vecs = vecs / norm.clamp(min=1e-5).sqrt()
        return vecs

    def message(self, edge_index, s_i, s_j, fm_i, fm_j, edge_attr=None):
        """
        Parameters
        ----------
        edge_index : torch.Tensor
            Edge index tensor containing the indices of the source and target nodes, shape (2, num_edges).
        s_i : torch.Tensor
            Scalar features of the source nodes, shape (num_edges, num_scalars).
        s_j : torch.Tensor
            Scalar features of the target nodes, shape (num_edges, num_scalars).
        fm_i : torch.Tensor
            Fourmomentum of the source nodes, shape (num_edges, in_vectors*4).
        fm_j : torch.Tensor
            Fourmomentum of the target nodes, shape (num_edges, in_vectors*4).
        edge_attr : torch.Tensor, optional
            Edge attributes tensor. If None, no edge attributes will be used, shape (num_edges, num_edge_attributes).

        Returns
        -------
        torch.Tensor
            Tensor of shape (num_edges, out_vectors*4) containing the predicted vectors for each edge.
        """
        fm_rel = self.operation(fm_i, fm_j)
        # should not be used with operation "single"
        if self.fm_norm:
            fm_rel_norm = lorentz_squarednorm(fm_rel).unsqueeze(-1)
            fm_rel_norm = fm_rel_norm.abs().sqrt().clamp(min=1e-6)
        else:
            fm_rel_norm = 1.0

        prefactor = torch.cat([s_i, s_j], dim=-1)
        if edge_attr is not None:
            prefactor = torch.cat([prefactor, edge_attr], dim=-1)
        prefactor = self.mlp(prefactor)
        prefactor = self.nonlinearity(prefactor, batch=edge_index[0])
        fm_rel = (fm_rel / fm_rel_norm)[:, None, :4]
        prefactor = prefactor.unsqueeze(-1)
        out = prefactor * fm_rel
        out = out.reshape(out.shape[0], -1)
        return out

    def get_operation(self, operation):
        """
        Parameters
        ----------
        operation : str
            Operation to perform on the fourmomenta. Options are "add", "diff", or "single".

        Returns
        -------
        callable
            A function that performs the specified operation on two fourmomenta tensors.
        """
        if operation == "diff":
            return torch.sub
        elif operation == "add":
            return torch.add
        elif operation == "single":
            return lambda fm_i, fm_j: fm_j
        else:
            raise ValueError(
                f"Invalid operation {operation}. Options are (add, diff, single)."
            )

    def get_nonlinearity(self, nonlinearity):
        """
        Parameters
        ----------
        nonlinearity : str or None
            Nonlinearity to apply to the output of the MLP. Options are None, "exp", "softplus", "softmax".

        Returns
        -------
        callable
            A function that applies the specified nonlinearity to the input tensor.
        """
        if nonlinearity == None:
            return lambda x, batch: x
        elif nonlinearity == "exp":
            return lambda x, batch: torch.clamp(x, min=-10, max=10).exp()
        elif nonlinearity == "softplus":
            return lambda x, batch: torch.nn.functional.softplus(x)
        elif nonlinearity == "softmax":

            def func(x, batch):
                ptr = get_ptr_from_batch(batch)
                return softmax(x, ptr=ptr)

            return func
        elif nonlinearity == "softmax_safe":

            def func(x, batch):
                ptr = get_ptr_from_batch(batch)
                return safe_softmax(x, ptr=ptr)

            return func
        else:
            raise ValueError(
                f"Invalid nonlinearity {nonlinearity}. Options are (None, exp, softplus, softmax)."
            )


class EquiMLP(EquiVectors):
    def __init__(
        self,
        n_vectors,
        num_blocks,
        *args,
        hidden_vectors=1,
        **kwargs,
    ):
        """
        Parameters
        ----------
        n_vectors : int
            Number of output vectors per particle.
        num_blocks : int
            Number of EquiEdgeConv blocks to use in the network.
        hidden_vectors : int, optional
            Number of hidden vectors in each EquiEdgeConv block. Default is 1.
        *args
        **kwargs
        """
        super().__init__()

        assert num_blocks >= 1
        in_vectors = [1] + [hidden_vectors] * (num_blocks - 1)
        out_vectors = [hidden_vectors] * (num_blocks - 1) + [n_vectors]
        self.blocks = nn.ModuleList(
            [
                EquiEdgeConv(
                    in_vectors=in_vectors[i],
                    out_vectors=out_vectors[i],
                    *args,
                    **kwargs,
                )
                for i in range(num_blocks)
            ]
        )

    def forward(self, fourmomenta, scalars=None, ptr=None):
        """
        Parameters
        ----------
        fourmomenta : torch.Tensor
            Tensor of shape (..., 4) containing the fourmomenta of the particles.
        scalars : torch.Tensor, optional
            Tensor of shape (..., num_scalars) containing scalar features for each particle. If None, a tensor of zeros will be created.
        ptr : torch.Tensor, optional
            Pointer tensor indicating the start and end of each batch for sparse tensors.

        Returns
        -------
        torch.Tensor
            Tensor of shape (..., n_vectors, 4) containing the predicted vectors for each particle.
        """
        # get edge_index and batch from ptr
        in_shape = fourmomenta.shape[:-1]
        if scalars is None:
            scalars = torch.zeros_like(fourmomenta[..., []])
        if len(in_shape) > 1:
            assert ptr is None, "ptr only supported for sparse tensors"
            edge_index, batch = build_edge_index_fully_connected(fourmomenta)
            fourmomenta = fourmomenta.reshape(math.prod(in_shape), 4)
            scalars = scalars.reshape(math.prod(in_shape), scalars.shape[-1])
        else:
            if ptr is None:
                # assume batch contains only one particle
                ptr = torch.tensor([0, len(fourmomenta)], device=fourmomenta.device)
            edge_index = get_edge_index_from_ptr(ptr)
            batch = None

        # pass through blocks
        for block in self.blocks:
            fourmomenta = block(
                fourmomenta, scalars=scalars, edge_index=edge_index, batch=batch
            )
        fourmomenta = fourmomenta.reshape(*in_shape, -1, 4)
        return fourmomenta


def safe_softmax(x, ptr):
    """Custom softmax implementation to control numerics."""
    seg_id = torch.arange(ptr.numel() - 1, device=x.device).repeat_interleave(
        ptr[1:] - ptr[:-1]
    )

    # rescale argument to avoid exp(large number)
    seg_max = scatter(x, seg_id, reduce="max")[seg_id].detach()
    z = x - seg_max

    # clamp to avoid rounding small values to zero (causes 'DivBackward0 returns nan')
    # this step is not included in standard softmax implementations
    z = z.clamp(min=-10)  # -10 works well; -20 and -5 already worse

    # actual softmax
    num = z.exp()
    den = scatter(num, seg_id, reduce="sum")[seg_id]
    out = num / den
    return out



import torch
import pytest
from tests.constants import TOLERANCES, MILD_TOLERANCES, BATCH_DIMS
from tests.helpers import lorentz_test

from lloca.utils.rand_transforms import (
    rand_lorentz,
    rand_rotation,
    rand_xyrotation,
    rand_ztransform,
    rand_boost,
)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("std_eta", [0.1, 1, 2])
@pytest.mark.parametrize(
    "transform_type",
    [
        rand_lorentz,
        rand_rotation,
        rand_xyrotation,
        rand_ztransform,
        rand_boost,
    ],
)
def test_rand_lorentz(batch_dims, std_eta, transform_type):
    dtype = torch.float64

    # collect N different kinds of transformations
    kwargs = {
        "shape": batch_dims,
        "dtype": dtype,
    }
    if transform_type in [rand_lorentz]:
        kwargs["std_eta"] = std_eta
    transform = transform_type(**kwargs)

    # test that this is a valid Lorentz transform
    lorentz_test(transform, **MILD_TOLERANCES)

    # test specific properties of the transform
    if transform_type in [rand_rotation, rand_xyrotation]:
        should_zero = torch.cat(
            [transform[..., 0, 1:].flatten(), transform[..., 1:, 0].flatten()]
        )
        if transform_type == rand_xyrotation:
            should_zero = torch.cat(
                [
                    should_zero,
                    transform[..., 3, 1:2].flatten(),
                    transform[..., 1:2, 3].flatten(),
                ]
            )
        torch.testing.assert_close(
            should_zero, torch.zeros_like(should_zero), **TOLERANCES
        )
